from collections.abc import Iterable
import math
from ...sampling.halton.prime import generate_n_primes


def halton(index: int, base: int) -> float:
    fraction: float = 1.0
    result: float = 0

    while index > 0:
        fraction = fraction / base
        result += fraction * (index % base)
        index = math.floor(index / base)

    return result


class HaltonSampleGenerator:
    index: int
    maxIndex: int
    primes: list[int]

    def __init__(self, count: int, offset: int, num_parameters: int):
        assert (count > 0)
        assert (offset >= 0)
        self.index = offset + 1
        self.maxIndex = offset + count + 1
        self.primes = generate_n_primes(num_parameters)

    def __iter__(self) -> Iterable[tuple[int, list[float]]]:
        return self

    def __next__(self) -> tuple[int, list[float]]:
        if self.index >= self.maxIndex:
            raise StopIteration

        i = 0
        values: list[float] = []
        for base in self.primes:
            values.append(halton(self.index, base))

        result: tuple[int, list[float]] = (self.index, values)
        self.index = self.index + 1
        return result




from customtkinter import CTkScrollableFrame
from customtkinter import CTkFrame
from customtkinter import CTkLabel
from customtkinter import CTkButton
from customtkinter import CTkEntry
import tkinter as tk

class OptimalParameterView(CTkScrollableFrame):
    def __init__(self, *args,
                 option_manager: None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        
        self.option_manager = option_manager
        
        self.render()

    def clear(self):
        self.containerFrame.destroy()
        
    def render(self):
        row = 0
        
        self.containerFrame = CTkFrame(self)
        self.containerFrame.grid(row=0, column=0, padx=(5, 5), pady=(5, 5), sticky="nsew")
        self.containerFrame.grid_columnconfigure((0, 1), weight=1)
        
        self.mode = self.option_manager.get_mode()
        self.steps = self.option_manager.get_steps()

        for step in self.steps:
            
            if self.mode != "Sampling: Random" and self.mode != "Sampling: Halton":
                name = step['name'].get()
                CTkLabel(self.containerFrame, text=name).grid(row=row, column=0, columnspan=1, padx=5, pady=5, sticky="")
                row += 1
            
            for param in step['parameter_objects']:
                CTkEntry(self.containerFrame, textvariable=param['name']).grid(row=row, column=0, padx=(5, 5), pady=(5, 5), sticky="ew")
            
                bb = CTkEntry(self.containerFrame)
                bb.grid(row=row, column=1, padx=(5, 5), pady=(5, 5), sticky="ew")
                bb.configure(textvariable=param['optimal_value'])
                row += 1
                
            if self.mode != "Optimization":
                break


from collections.abc import Iterable
import math
import asyncio
import concurrent
import datetime
from ..utils import utils
from ..sampling.halton.halton import HaltonSampleGenerator
from ..sampling.random.random_sampler import RandomSampleGenerator
from ..sampling.sampler_task import SamplerTask
from ..sampling.sample_trace_writer import SampleTraceWriter


def weighted_value(weight: float, lower: float, upper: float) -> float:
    return lower + weight * (upper - lower)


def get_static_parameters(args: dict[str, any]) -> dict[str, any]:
    static_parameters: dict[str, any] = {}
    for param in args["param"]:
        static_parameters[param["name"]] = param["value"]
    return static_parameters


def get_objective_names(objfunc: dict[str, any]) -> list[str]:
    objective_names: list[str] = []
    for of in objfunc:
        objective_names.append(of["name"])
    return objective_names


def thread_function(task: SamplerTask) -> tuple[bool, SamplerTask]:
    return task.run_task(), task


def create_generator(method: str, count: int, num_parameters: int, **kwargs) -> Iterable[tuple[int, list[float]]]:
    if method == "halton":
        offset: int = 0
        if "offset" in kwargs:
            offset = kwargs["offset"]
        return HaltonSampleGenerator(count, offset, num_parameters)
    elif method == "random":
        return RandomSampleGenerator(count, num_parameters)

    raise Exception("Sampling method is not recognized")


def run_sampler(steps: list[dict[str, any]], args: dict[str, any], count: int, num_threads: int, method: str = "halton",
                metainfo: dict[str, any] = None, conf: dict[str, any] = None, trace_file: str = "trace.csv",
                **kwargs) -> dict[int, tuple[dict[str, any], dict[str, any]]]:
    param_names, bounds, objfunc = utils.get_step_info(steps, 0)
    generator: Iterable[tuple[int, list[float]]] = create_generator(method, count, len(param_names), **kwargs)
    objective_names: list[str] = get_objective_names(objfunc)
    static_parameters: dict[str, any] = get_static_parameters(args)
    url: str = args["url"]
    files: list[str] = args["files"]

    trace: dict[int, tuple[dict[str, float], dict[str, float]]] = {}
    trace_writer: SampleTraceWriter = SampleTraceWriter(trace_file)
    trace_writer.write_header(param_names, objective_names)

    with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
        futures = []
        for sample_id, sample in generator:
            params: dict[str, float] = {}
            index: int = 0
            while index < len(sample):
                params[param_names[index]] = weighted_value(sample[index], bounds[0][index], bounds[1][index])
                index += 1

            task: SamplerTask = SamplerTask(sample_id, params, objfunc, static_parameters, url, files, metainfo, conf)
            futures.append(executor.submit(thread_function, task))
        # for future in concurrent.futures.as_completed(futures):
        #     pass
        num_finished: int = 0
        percentage: float
        last_percentage: float = 0
        for future in concurrent.futures.as_completed(futures):
            try:
                successful, task = future.result()

                if successful:
                    trace[task.task_id] = (task.parameters, task.result)
                    trace_writer.append_sample(task.task_id, task.parameters, task.result)
                else:
                    print("Failed to successfully execute task: {}", task.task_id, flush=True)
            except asyncio.CancelledError as ce:
                pass
            except asyncio.InvalidStateError as ise:
                pass
            except Exception as ex:
                print(ex, flush=True)

            num_finished = num_finished + 1
            percentage = math.trunc(num_finished / count * 1000) / 10
            if percentage > last_percentage:
                last_percentage = percentage
                print("{}% Done {}".format(percentage, datetime.datetime.now()), flush=True)

    return trace



from typing import Union, Tuple, Optional

from customtkinter import CTkLabel
from customtkinter import CTkButton
from customtkinter import CTkEntry
from customtkinter import CTkInputDialog
from customtkinter import CTkTextbox

class NoticeWindow(CTkInputDialog):
    """
    Dialog with extra window, message, entry widget, cancel and ok button.
    For detailed information check out the documentation.
    """

    def __init__(self, *args,
                 message: None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        
        self.geometry("800x800")

        self.message = message

    def _create_widgets(self):

        self.grid_columnconfigure(0, weight=1)
        self.rowconfigure(0, weight=1)

        self.textbox = CTkTextbox(self, state="normal")
        self.textbox.grid(row=0, column=0, padx=(20, 20), pady=(20, 20), sticky="nsew")
        self.textbox.insert("0.0", self.message)
        self.textbox.configure(wrap="word")
        self.textbox.configure(state="disabled")
            
        self._ok_button = CTkButton(master=self,
                                    width=100,
                                    border_width=0,
                                    fg_color=self._button_fg_color,
                                    hover_color=self._button_hover_color,
                                    text_color=self._button_text_color,
                                    text='OK',
                                    command=self._ok_event)
        self._ok_button.grid(row=1, column=0, columnspan=1, padx=(20, 10), pady=(0, 20), sticky="ew")
        
    def _ok_event(self, event=None):
        self.grab_release()
        self.destroy()

    def _on_closing(self):
        self.grab_release()
        self.destroy()

    def _cancel_event(self):
        self.grab_release()
        self.destroy()